"""Define Home Assistant-related functionality."""
import argparse
from dataclasses import dataclass
from typing import Any, Dict, Optional, Union

from ecowitt2mqtt.const import (
    DATA_POINT_CO2,
    DATA_POINT_DEWPOINT,
    DATA_POINT_FEELSLIKE,
    DATA_POINT_GLOB_BAROM,
    DATA_POINT_GLOB_BATT,
    DATA_POINT_GLOB_BATT_BINARY,
    DATA_POINT_GLOB_GUST,
    DATA_POINT_GLOB_HUMIDITY,
    DATA_POINT_GLOB_MOISTURE,
    DATA_POINT_GLOB_RAIN,
    DATA_POINT_GLOB_TEMP,
    DATA_POINT_GLOB_WIND,
    DATA_POINT_HEATINDEX,
    DATA_POINT_PM25,
    DATA_POINT_PM25_24H,
    DATA_POINT_SOLARRADIATION,
    DATA_POINT_SOLARRADIATION_LUX,
    DATA_POINT_SOLARRADIATION_PERCEIVED,
    DATA_POINT_UV,
    DATA_POINT_WINDCHILL,
    DATA_POINT_WINDDIR,
    LOGGER,
    UNIT_SYSTEM_IMPERIAL,
    UNIT_SYSTEM_METRIC,
)
from ecowitt2mqtt.device import Device

PLATFORM_BINARY_SENSOR = "binary_sensor"
PLATFORM_SENSOR = "sensor"

DEVICE_CLASS_BATTERY = "battery"
DEVICE_CLASS_CO = "carbon_monoxide"
DEVICE_CLASS_HUMIDITY = "humidity"
DEVICE_CLASS_ILLUMINANCE = "illuminance"
DEVICE_CLASS_PM25 = "pm25"
DEVICE_CLASS_PRESSURE = "pressure"
DEVICE_CLASS_TEMPERATURE = "temperature"

UNIT_CLASS_PRESSURE = "pressure"
UNIT_CLASS_RAIN = "rain"
UNIT_CLASS_TEMPERATURE = "temperature"
UNIT_CLASS_WIND = "wind"


@dataclass
class EntityDescription:
    """Define a description (set of characteristics) of a Home Assistant entity."""

    platform: str

    device_class: Optional[str] = None
    icon: Optional[str] = None
    unit: Optional[str] = None
    unit_class: Optional[str] = None


ENTITY_DESCRIPTIONS = {
    DATA_POINT_GLOB_BAROM: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_PRESSURE,
        unit_class=UNIT_CLASS_PRESSURE,
    ),
    DATA_POINT_GLOB_BATT: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_BATTERY,
        unit="v",
    ),
    DATA_POINT_GLOB_BATT_BINARY: EntityDescription(
        platform=PLATFORM_BINARY_SENSOR,
        device_class=DEVICE_CLASS_BATTERY,
    ),
    DATA_POINT_GLOB_GUST: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:weather-windy",
        unit_class=UNIT_CLASS_WIND,
    ),
    DATA_POINT_GLOB_HUMIDITY: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_HUMIDITY,
        unit="%",
    ),
    DATA_POINT_GLOB_MOISTURE: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:water-percent",
        unit="%",
    ),
    DATA_POINT_GLOB_RAIN: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:water",
        unit_class=UNIT_CLASS_RAIN,
    ),
    DATA_POINT_GLOB_TEMP: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_TEMPERATURE,
        unit_class=UNIT_CLASS_TEMPERATURE,
    ),
    DATA_POINT_GLOB_WIND: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:weather-windy",
        unit_class=UNIT_CLASS_WIND,
    ),
    DATA_POINT_CO2: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_CO,
        unit="ppm",
    ),
    DATA_POINT_DEWPOINT: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:thermometer",
        unit_class=UNIT_CLASS_TEMPERATURE,
    ),
    DATA_POINT_FEELSLIKE: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:thermometer",
        unit_class=UNIT_CLASS_TEMPERATURE,
    ),
    DATA_POINT_HEATINDEX: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:thermometer",
        unit_class=UNIT_CLASS_TEMPERATURE,
    ),
    DATA_POINT_PM25: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_PM25,
        unit="µg/m^3",
    ),
    DATA_POINT_PM25_24H: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_PM25,
        unit="µg/m^3",
    ),
    DATA_POINT_SOLARRADIATION: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_ILLUMINANCE,
        unit="w/m^2",
    ),
    DATA_POINT_SOLARRADIATION_LUX: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_ILLUMINANCE,
        unit="lx",
    ),
    DATA_POINT_SOLARRADIATION_PERCEIVED: EntityDescription(
        platform=PLATFORM_SENSOR,
        device_class=DEVICE_CLASS_ILLUMINANCE,
        unit="%",
    ),
    DATA_POINT_UV: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:weather-sunny",
        unit="index",
    ),
    DATA_POINT_WINDCHILL: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:weather-windy",
        unit_class=UNIT_CLASS_TEMPERATURE,
    ),
    DATA_POINT_WINDDIR: EntityDescription(
        platform=PLATFORM_SENSOR,
        icon="mdi:weather-windy",
        unit="°",
    ),
}

UNIT_MAPPING = {
    UNIT_CLASS_PRESSURE: {UNIT_SYSTEM_IMPERIAL: "inHg", UNIT_SYSTEM_METRIC: "hPa"},
    UNIT_CLASS_RAIN: {UNIT_SYSTEM_IMPERIAL: "in", UNIT_SYSTEM_METRIC: "mm"},
    UNIT_CLASS_TEMPERATURE: {UNIT_SYSTEM_IMPERIAL: "°F", UNIT_SYSTEM_METRIC: "°C"},
    UNIT_CLASS_WIND: {UNIT_SYSTEM_IMPERIAL: "mph", UNIT_SYSTEM_METRIC: "km/h"},
}


def get_entity_description(
    key: str, value: Union[float, int, str]
) -> EntityDescription:
    """Get an entity description for a data key.

    1. Return a specific data point if it exists.
    2. Return a globbed data point if it exists.
    3. Return defaults if no specific or globbed data points exist.
    """
    if DATA_POINT_GLOB_BATT in key and isinstance(value, str):
        # Because Ecowitt doesn't give us a clear way to know what sort of battery
        # we're looking at (a binary on/off battery or one that reports voltage), we
        # check its value: strings are binary, floats are voltage:
        return ENTITY_DESCRIPTIONS[DATA_POINT_GLOB_BATT_BINARY]

    if key in ENTITY_DESCRIPTIONS:
        return ENTITY_DESCRIPTIONS[key]

    globbed_descriptions = [v for k, v in ENTITY_DESCRIPTIONS.items() if k in key]
    if globbed_descriptions:
        return globbed_descriptions[0]

    LOGGER.info("No entity description found for key: %s", key)
    return EntityDescription(platform=PLATFORM_SENSOR)


class HassDiscovery:
    """Define a Home Assistant MQTT Discovery manager."""

    def __init__(self, device: Device, args: argparse.Namespace) -> None:
        """Initialize."""
        self._args = args
        self._config_payloads: Dict[str, Dict[str, Any]] = {}
        self._device = device

    def _get_topic(self, key: str, platform: str, topic_type: str) -> str:
        """Get the attributes topic for a particular entity type."""
        return (
            f"{self._args.hass_discovery_prefix}/{platform}/{self._device.unique_id}/"
            f"{key}/{topic_type}"
        )

    def get_config_payload(
        self, key: str, value: Union[float, int, str]
    ) -> Dict[str, Any]:
        """Return the config payload for a particular entity type."""
        if key in self._config_payloads:
            return self._config_payloads[key]

        description = get_entity_description(key, value)

        if description.unit_class:
            description.unit = UNIT_MAPPING[description.unit_class][
                self._args.output_unit_system
            ]

        self._config_payloads[key] = {
            "availability_topic": self._get_topic(
                key, description.platform, "availability"
            ),
            "device": {
                "identifiers": [self._device.unique_id],
                "manufacturer": self._device.manufacturer,
                "model": self._device.name,
                "name": self._device.name,
                "sw_version": self._device.station_type,
            },
            "name": key,
            "qos": 1,
            "state_topic": self._get_topic(key, description.platform, "state"),
            "unique_id": f"{self._device.unique_id}_{key}",
        }

        if description.device_class:
            self._config_payloads[key]["device_class"] = description.device_class
        if description.icon:
            self._config_payloads[key]["icon"] = description.icon
        if description.unit:
            self._config_payloads[key]["native_unit_of_measurement"] = description.unit

        return self._config_payloads[key]

    def get_config_topic(self, key: str, value: Union[float, int, str]) -> str:
        """Return the config topic for a particular entity type."""
        description = get_entity_description(key, value)
        return self._get_topic(key, description.platform, "config")



"""Define battery utilities."""
from typing import Union

BATTERY_STATE_OFF = "off"
BATTERY_STATE_ON = "on"


def calculate_battery(value: Union[float, int]) -> Union[float, str]:
    """Calculate a battery value.

    1. If the value is a float, we assume it represents voltage and return it as-is.
    2. If the value is an int, we assume it represents a binary state:
         * 0: OK
         * 1: Low
    """
    if isinstance(value, float):
        return value

    if value == 0:
        return BATTERY_STATE_OFF
    return BATTERY_STATE_ON



"""Define aiohttp routes."""
import asyncio
import json
from typing import Any, Dict, Union

from aiohttp import web
from asyncio_mqtt import Client, MqttError

from ecowitt2mqtt.const import LOGGER
from ecowitt2mqtt.data import DataProcessor
from ecowitt2mqtt.hass import HassDiscovery


def _generate_payload(data: Union[Dict[str, Any], float, str]) -> bytes:
    """Generate a binary MQTT payload from input data."""
    if isinstance(data, dict):
        return json.dumps(data).encode("utf-8")

    if isinstance(data, str):
        return data.encode("utf-8")

    return str(data).encode("utf-8")


async def _async_publish_to_hass_discovery(
    client: Client, data: Dict[str, Any], discovery_manager: HassDiscovery
) -> None:
    """Publish data to appropriate topics for Home Assistant Discovery."""
    LOGGER.debug("Publishing according to Home Assistant MQTT Discovery standard")

    try:
        async with client:
            tasks = []
            for key, value in data.items():
                config_payload = discovery_manager.get_config_payload(key, value)
                config_topic = discovery_manager.get_config_topic(key, value)

                tasks.append(
                    client.publish(config_topic, _generate_payload(config_payload))
                )
                tasks.append(
                    client.publish(
                        config_payload["availability_topic"],
                        _generate_payload("online"),
                    )
                )
                tasks.append(
                    client.publish(
                        config_payload["state_topic"],
                        _generate_payload(value),
                        retain=True,
                    )
                )

            await asyncio.gather(*tasks)
    except MqttError as err:
        LOGGER.error("Error while publishing to HASS Discovery: %s", err)
        return

    LOGGER.info("Published to HASS discovery: %s", data)


async def _async_publish_to_topic(
    client: Client, data: Dict[str, Any], topic: str
) -> None:
    """Publish data to a single MQTT topic."""
    LOGGER.debug("Publishing entire device payload to single topic: %s", topic)

    try:
        async with client:
            await client.publish(topic, _generate_payload(data))
    except MqttError as err:
        LOGGER.error("Error while publishing to %s: %s", topic, err)
        return

    LOGGER.info("Published to %s: %s", topic, data)


async def async_publish_payload(request: web.Request) -> None:
    """Define the endpoint for the Ecowitt device to post data to."""
    args = request.app["args"]

    payload = dict(await request.post())
    LOGGER.debug("Received data from Ecowitt device: %s", payload)

    data_processor = DataProcessor(payload, args)
    data = data_processor.generate_data()

    client = Client(
        args.mqtt_broker,
        port=args.mqtt_port,
        username=args.mqtt_username,
        password=args.mqtt_password,
        logger=LOGGER,
    )

    if args.hass_discovery:
        discovery_managers = request.app["hass_discovery_managers"]

        if data_processor.device.unique_id in discovery_managers:
            discovery_manager = discovery_managers[data_processor.device.unique_id]
        else:
            discovery_manager = HassDiscovery(data_processor.device, args)
            discovery_managers[data_processor.device.unique_id] = discovery_manager

        await _async_publish_to_hass_discovery(client, data, discovery_manager)
    else:
        if args.mqtt_topic:
            topic = args.mqtt_topic
        else:
            topic = f"ecowitt2mqtt/{data_processor.device.unique_id}"

        await _async_publish_to_topic(client, data, topic)



"""Define an Ecowitt device."""
from typing import Any, Dict, NamedTuple, Optional

from ecowitt2mqtt.const import LOGGER

DEFAULT_MANUFACTURER = "Unknown"
DEFAULT_NAME = "Unknown Device"
DEFAULT_STATION_TYPE = "Unknown Station Type"
DEFAULT_UNIQUE_ID = "default"

DEVICE_DATA = {
    "GW1000": ("Ecowitt", "GW1000"),
    "GW1100": ("Ecowitt", "GW1100"),
    "WH2650": ("Fine Offset", "WH2650"),
    "WS2900": ("Ambient Weather", "WS-2902C"),
}


class Device(NamedTuple):
    """Simple data object to provide device details."""

    unique_id: str
    manufacturer: str
    name: str
    station_type: Optional[str]


def get_device_from_raw_payload(payload: Dict[str, Any]) -> Device:
    """Return a device based upon a model string."""
    model = payload["model"]
    station_type = payload.get("stationtype", DEFAULT_STATION_TYPE)
    unique_id = payload.get("PASSKEY", DEFAULT_UNIQUE_ID)

    if model in DEVICE_DATA:
        manufacturer, name = DEVICE_DATA[model]
    else:
        matches = [v for k, v in DEVICE_DATA.items() if k in model]
        if matches:
            manufacturer, name = matches[0]
        else:
            LOGGER.info(
                (
                    "Unknown device; please report it at "
                    "https://github.com/bachya/ecowitt2mqtt (payload: %s)"
                ),
                payload,
            )
            manufacturer = DEFAULT_MANUFACTURER
            name = DEFAULT_NAME

    return Device(unique_id, manufacturer, name, station_type)



"""Define helpers to process data from an Ecowitt device."""
import argparse
from functools import partial
import inspect
from typing import Any, Callable, Dict, Optional, Union

from ecowitt2mqtt.const import (
    DATA_POINT_DEWPOINT,
    DATA_POINT_FEELSLIKE,
    DATA_POINT_GLOB_BAROM,
    DATA_POINT_GLOB_BATT,
    DATA_POINT_GLOB_RAIN,
    DATA_POINT_GLOB_TEMP,
    DATA_POINT_GLOB_WIND,
    DATA_POINT_HEATINDEX,
    DATA_POINT_HUMIDITY,
    DATA_POINT_SOLARRADIATION,
    DATA_POINT_SOLARRADIATION_LUX,
    DATA_POINT_SOLARRADIATION_PERCEIVED,
    DATA_POINT_TEMPF,
    DATA_POINT_WINDCHILL,
    DATA_POINT_WINDSPEEDMPH,
)
from ecowitt2mqtt.device import get_device_from_raw_payload
from ecowitt2mqtt.util.battery import calculate_battery
from ecowitt2mqtt.util.meteo import (
    calculate_dew_point,
    calculate_feels_like,
    calculate_heat_index,
    calculate_illuminance_wm2_to_lux,
    calculate_illuminance_wm2_to_perceived,
    calculate_pressure,
    calculate_rain_volume,
    calculate_temperature,
    calculate_wind_chill,
    calculate_wind_speed,
)

DEFAULT_KEYS_TO_IGNORE = ["PASSKEY", "dateutc", "freq", "model", "stationtype"]

CALCULATOR_FUNCTION_MAP: Dict[str, Callable] = {
    DATA_POINT_DEWPOINT: calculate_dew_point,
    DATA_POINT_FEELSLIKE: calculate_feels_like,
    DATA_POINT_GLOB_BAROM: calculate_pressure,
    DATA_POINT_GLOB_BATT: calculate_battery,
    DATA_POINT_GLOB_RAIN: calculate_rain_volume,
    DATA_POINT_GLOB_TEMP: calculate_temperature,
    DATA_POINT_GLOB_WIND: calculate_wind_speed,
    DATA_POINT_HEATINDEX: calculate_heat_index,
    DATA_POINT_SOLARRADIATION_LUX: calculate_illuminance_wm2_to_lux,
    DATA_POINT_SOLARRADIATION_PERCEIVED: calculate_illuminance_wm2_to_perceived,
    DATA_POINT_WINDCHILL: calculate_wind_chill,
}

DEW_POINT_KEYS = (DATA_POINT_TEMPF, DATA_POINT_HUMIDITY)
FEELS_LIKE_KEYS = (DATA_POINT_TEMPF, DATA_POINT_HUMIDITY, DATA_POINT_WINDSPEEDMPH)
HEAT_INDEX_KEYS = (DATA_POINT_TEMPF, DATA_POINT_HUMIDITY)
WIND_CHILL_KEYS = (DATA_POINT_TEMPF, DATA_POINT_WINDSPEEDMPH)
ILLUMINANCE_KEYS = (DATA_POINT_SOLARRADIATION,)


def de_unit_key(key: str) -> str:
    """Remove the unit from a key."""
    if key.endswith("f"):
        return key[:-1]
    if key.endswith("in"):
        return key[:-2]
    if key.endswith("mph"):
        return key[:-3]
    return key


def get_data_type(key: str) -> Optional[str]:
    """Get the data "type" (if it exists) for a specific data key."""
    if key in CALCULATOR_FUNCTION_MAP:
        return key

    matches = [k for k in CALCULATOR_FUNCTION_MAP if k in key]
    if matches:
        return matches[0]

    return None


def get_typed_value(value: str) -> Union[float, int, str]:
    """Take a string and return its properly typed counterpart."""
    if value.isdigit():
        # Integer:
        return int(value)

    try:
        # Float:
        return float(value)
    except ValueError:
        # String:
        return value


class DataProcessor:  # pylint: disable=too-few-public-methods
    """Define an object that holds processed payload data from the device."""

    def __init__(self, payload: Dict[str, Any], args: argparse.Namespace) -> None:
        """Initialize."""
        self._args = args
        self._input_unit_system = args.input_unit_system
        self._output_unit_system = args.output_unit_system

        self._payload: Dict[str, Union[float, str]] = {}
        for key, value in payload.items():
            self._payload[key] = get_typed_value(value)

        self.device = get_device_from_raw_payload(payload)

    def _get_calculator_func(
        self, key: str, *args: Union[float, str], **kwargs: str
    ) -> Optional[Callable]:
        """Get a data calculator function for a particular data key."""
        data_type = get_data_type(key)

        if not data_type:
            return None

        func = CALCULATOR_FUNCTION_MAP[data_type]

        func_params = inspect.signature(func).parameters
        if "input_unit_system" in func_params:
            kwargs["input_unit_system"] = self._input_unit_system
        if "output_unit_system" in func_params:
            kwargs["output_unit_system"] = self._output_unit_system

        return partial(func, *args, **kwargs)

    def generate_data(self) -> Dict[str, Union[float, str]]:
        """Generate a parsed data payload."""
        data: Dict[str, Any] = {}

        for target_key, value in self._payload.items():
            if target_key in DEFAULT_KEYS_TO_IGNORE:
                continue

            calculate = self._get_calculator_func(target_key, value)

            if self._args.raw_data or not calculate:
                data[target_key] = value
                continue

            output = calculate()
            target_key = de_unit_key(target_key)
            data[target_key] = output

        for target_key, input_keys in [
            (DATA_POINT_DEWPOINT, DEW_POINT_KEYS),
            (DATA_POINT_FEELSLIKE, FEELS_LIKE_KEYS),
            (DATA_POINT_HEATINDEX, HEAT_INDEX_KEYS),
            (DATA_POINT_SOLARRADIATION_LUX, ILLUMINANCE_KEYS),
            (DATA_POINT_SOLARRADIATION_PERCEIVED, ILLUMINANCE_KEYS),
            (DATA_POINT_WINDCHILL, WIND_CHILL_KEYS),
        ]:
            if not all(k in self._payload for k in input_keys):
                continue

            calculate = self._get_calculator_func(
                target_key, *[self._payload[k] for k in input_keys]
            )

            if not calculate:
                continue

            output = calculate()
            data[target_key] = output

        return data



"""Define meteorological utilities."""
import math
from typing import Optional, cast

import meteocalc

from ecowitt2mqtt.const import LOGGER, UNIT_SYSTEM_IMPERIAL


def _get_temperature_object(temperature: float, unit_system: str) -> meteocalc.Temp:
    """Get a meteocalc temperature object based on a temperature and unit system."""
    if unit_system == UNIT_SYSTEM_IMPERIAL:
        unit = "f"
    else:
        unit = "c"
    return meteocalc.Temp(temperature, unit)


def calculate_dew_point(
    temperature: float,
    humidity: float,
    *,
    input_unit_system: str = UNIT_SYSTEM_IMPERIAL,
    output_unit_system: str = UNIT_SYSTEM_IMPERIAL
) -> float:
    """Calculate dew point in the appropriate unit system."""
    temp_obj = _get_temperature_object(temperature, input_unit_system)
    dew_point_obj = meteocalc.dew_point(temp_obj, humidity)

    if output_unit_system == UNIT_SYSTEM_IMPERIAL:
        value = round(dew_point_obj.f, 1)
    else:
        value = round(dew_point_obj.c, 1)
    return cast(float, value)


def calculate_feels_like(
    temperature: float,
    humidity: float,
    wind_speed: float,
    *,
    input_unit_system: str = UNIT_SYSTEM_IMPERIAL,
    output_unit_system: str = UNIT_SYSTEM_IMPERIAL
) -> float:
    """Calculate "feels like" temperature in the appropriate unit system."""
    temp_obj = _get_temperature_object(temperature, input_unit_system)
    feels_like_obj = meteocalc.feels_like(temp_obj, humidity, wind_speed)

    if output_unit_system == UNIT_SYSTEM_IMPERIAL:
        value = round(feels_like_obj.f, 1)
    else:
        value = round(feels_like_obj.c, 1)
    return cast(float, value)


def calculate_heat_index(
    temperature: float,
    humidity: float,
    *,
    input_unit_system: str = UNIT_SYSTEM_IMPERIAL,
    output_unit_system: str = UNIT_SYSTEM_IMPERIAL
) -> float:
    """Calculate heat index in the appropriate unit system."""
    temp_obj = _get_temperature_object(temperature, input_unit_system)
    heat_index_obj = meteocalc.heat_index(temp_obj, humidity)

    if output_unit_system == UNIT_SYSTEM_IMPERIAL:
        value = round(heat_index_obj.f, 1)
    else:
        value = round(heat_index_obj.c, 1)
    return cast(float, value)


def calculate_illuminance_wm2_to_lux(value: float) -> float:
    """Calculate illuminance (in lux)."""
    return round(float(value) / 0.0079, 1)


def calculate_illuminance_wm2_to_perceived(value: float) -> float:
    """Calculate illuminance (in lux)."""
    lux = calculate_illuminance_wm2_to_lux(value)
    try:
        perceived = round(math.log10(lux) / 5, 2) * 100
    except ValueError:
        # If we've approached negative infinity, we'll get a math domain error; in that
        # case, return 0.0:
        return 0.0
    if perceived < 0:
        return 0.0
    return perceived


def calculate_pressure(
    value: float,
    *,
    input_unit_system: str = UNIT_SYSTEM_IMPERIAL,
    output_unit_system: str = UNIT_SYSTEM_IMPERIAL
) -> float:
    """Calculate pressure in the appropriate unit system."""
    if input_unit_system == output_unit_system:
        return value
    if output_unit_system == UNIT_SYSTEM_IMPERIAL:
        return round(value / 33.8639, 3)
    return round(value * 33.8639, 3)


def calculate_rain_volume(
    value: float,
    *,
    input_unit_system: str = UNIT_SYSTEM_IMPERIAL,
    output_unit_system: str = UNIT_SYSTEM_IMPERIAL
) -> float:
    """Calculate rain volume in the appropriate unit system."""
    if input_unit_system == output_unit_system:
        return value
    if output_unit_system == UNIT_SYSTEM_IMPERIAL:
        return round(value / 25.4, 1)
    return round(value * 25.4, 1)


def calculate_temperature(
    value: float,
    *,
    input_unit_system: str = UNIT_SYSTEM_IMPERIAL,
    output_unit_system: str = UNIT_SYSTEM_IMPERIAL
) -> float:
    """Calculate temperature in the appropriate unit system."""
    temp_obj = _get_temperature_object(value, input_unit_system)

    if output_unit_system == UNIT_SYSTEM_IMPERIAL:
        value = round(temp_obj.f, 1)
    else:
        value = round(temp_obj.c, 1)
    return value


def calculate_wind_chill(
    temperature: float,
    wind_speed: float,
    *,
    input_unit_system: str = UNIT_SYSTEM_IMPERIAL,
    output_unit_system: str = UNIT_SYSTEM_IMPERIAL
) -> Optional[float]:
    """Calculate wind chill in the appropriate unit system."""
    temp_obj = _get_temperature_object(temperature, input_unit_system)

    try:
        wind_chill_obj = meteocalc.wind_chill(temp_obj, wind_speed)
    except ValueError as err:
        LOGGER.debug("%s (temperature: %s, wind speed: %s)", err, temp_obj, wind_speed)
        return None

    if output_unit_system == UNIT_SYSTEM_IMPERIAL:
        value = round(wind_chill_obj.f, 1)
    else:
        value = round(wind_chill_obj.c, 1)
    return cast(float, value)


def calculate_wind_speed(
    value: float,
    *,
    input_unit_system: str = UNIT_SYSTEM_IMPERIAL,
    output_unit_system: str = UNIT_SYSTEM_IMPERIAL
) -> float:
    """Calculate wind speed in the appropriate unit system."""
    if input_unit_system == output_unit_system:
        return value
    if output_unit_system == UNIT_SYSTEM_IMPERIAL:
        return round(value / 1.60934, 1)
    return round(value * 1.60934, 1)



"""Define the main application."""
import argparse
import logging

from aiohttp import web

from ecowitt2mqtt.const import LOGGER, UNIT_SYSTEM_IMPERIAL
from ecowitt2mqtt.mqtt import async_publish_payload

DEFAULT_AIOHTTP_ENDPOINT = "/data/report"
DEFAULT_AIOHTTP_PORT = 8080
DEFAULT_HASS_DISCOVERY_PREFIX = "homeassistant"
DEFAULT_LOG_LEVEL_STRING = "INFO"
DEFAULT_MQTT_PORT = 1883


def get_arguments() -> argparse.Namespace:
    """Get passed-in arguments."""
    parser = argparse.ArgumentParser(
        description="Send data from Ecowitt devices to an MQTT broker"
    )

    parser.add_argument(
        "-l",
        "--log-level",
        action="store",
        default=DEFAULT_LOG_LEVEL_STRING,
        help=f"The logging level (default: {DEFAULT_LOG_LEVEL_STRING})",
    )

    # MQTT
    parser.add_argument(
        "--mqtt-broker",
        action="store",
        required=True,
        type=str,
        help="The hostname or IP address of the MQTT broker",
    )
    parser.add_argument(
        "--mqtt-port",
        action="store",
        default=DEFAULT_MQTT_PORT,
        type=int,
        help=f"The port of the MQTT broker (default: {DEFAULT_MQTT_PORT})",
    )
    parser.add_argument(
        "--mqtt-username",
        action="store",
        default=None,
        type=str,
        help="The username to use with the MQTT broker (default: None)",
    )
    parser.add_argument(
        "--mqtt-password",
        action="store",
        default=None,
        type=str,
        help="The password to use with the MQTT broker (default: None)",
    )
    parser.add_argument(
        "--mqtt-topic",
        action="store",
        type=str,
        help="The MQTT topic to publish the device's data to (default: ecowitt2mqtt/<ID>)",
    )

    # Home Assistant MQTT Discovery
    parser.add_argument(
        "--hass-discovery",
        action="store_const",
        const=True,
        help="Publish data in the Home Assistant MQTT Discovery format",
    )
    parser.add_argument(
        "--hass-discovery-prefix",
        default=DEFAULT_HASS_DISCOVERY_PREFIX,
        help=(
            "The Home Assistant discovery prefix to use "
            f"(default: {DEFAULT_HASS_DISCOVERY_PREFIX})"
        ),
    )

    # Web Server
    parser.add_argument(
        "--endpoint",
        action="store",
        default=DEFAULT_AIOHTTP_ENDPOINT,
        type=str,
        help=(
            "The relative endpoint/path to serve the web app on "
            f"(default: {DEFAULT_AIOHTTP_ENDPOINT})"
        ),
    )
    parser.add_argument(
        "--port",
        action="store",
        default=DEFAULT_AIOHTTP_PORT,
        type=int,
        help=f"The port to serve the web app on (default: {DEFAULT_AIOHTTP_PORT})",
    )

    # Misc.
    parser.add_argument(
        "--raw-data",
        action="store_const",
        const=True,
        help="Return raw data (don't attempt to translate any values)",
    )
    parser.add_argument(
        "--input-unit-system",
        action="store",
        default=UNIT_SYSTEM_IMPERIAL,
        type=str,
        help=f"The input unit system used by the device (default: {UNIT_SYSTEM_IMPERIAL})",
    )
    parser.add_argument(
        "--output-unit-system",
        action="store",
        default=UNIT_SYSTEM_IMPERIAL,
        type=str,
        help=f"The unit system to use in output (default: {UNIT_SYSTEM_IMPERIAL})",
    )

    return parser.parse_args()


def main() -> None:
    """Run."""
    args = get_arguments()
    logging.basicConfig(level=getattr(logging, args.log_level))

    LOGGER.debug("Using arguments: %s", args)

    app = web.Application()
    app.add_routes([web.post(args.endpoint, async_publish_payload)])  # type: ignore

    app["args"] = args
    app["hass_discovery_managers"] = {}

    web.run_app(app, port=args.port)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ecowitt2mqtt', 'ecowitt2mqtt.util']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'asyncio-mqtt>=0.7.0,<0.8.0',
 'meteocalc>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['ecowitt2mqtt = ecowitt2mqtt.main:main']}

setup_kwargs = {
    'name': 'ecowitt2mqtt',
    'version': '2.0.1',
    'description': 'A small web server to send data from Ecowitt devices to an MQTT Broker',
    'long_description': '# 🔘 ecowitt2mqtt: Send Ecowitt device data to an MQTT broker\n\n[![PyPi](https://img.shields.io/pypi/v/ecowitt2mqtt.svg)](https://pypi.python.org/pypi/ecowitt2mqtt)\n[![Docker Hub](https://img.shields.io/docker/pulls/bachya/ecowitt2mqtt)](https://hub.docker.com/r/bachya/ecowitt2mqtt)\n[![Version](https://img.shields.io/pypi/pyversions/ecowitt2mqtt.svg)](https://pypi.python.org/pypi/ecowitt2mqtt)\n[![License](https://img.shields.io/pypi/l/ecowitt2mqtt.svg)](https://github.com/bachya/ecowitt2mqtt/blob/master/LICENSE)\n[![Maintainability](https://api.codeclimate.com/v1/badges/a03c9e96f19a3dc37f98/maintainability)](https://codeclimate.com/github/bachya/ecowitt2mqtt/maintainability)\n[![Say Thanks](https://img.shields.io/badge/SayThanks-!-1EAEDB.svg)](https://saythanks.io/to/bachya)\n\n`ecowitt2mqtt` is a small CLI/web server that allows [Ecowitt](http://www.ecowitt.com)\ndevice data to be sent to an MQTT broker.\n\n- [Installation](#installation)\n- [Python Versions](#python-versions)\n- [Quick Start](#quick-start)\n- [Advanced Usage](#advanced-usage)\n- [Contributing](#contributing)\n\n# Installation\n\n```python\npip install ecowitt2mqtt\n```\n\n# Python Versions\n\n`ecowitt2mqtt` is currently supported on:\n\n* Python 3.6\n* Python 3.7\n* Python 3.8 \n\n# Quick Start\n\nNote that this README assumes:\n\n* you have access to an MQTT broker\n* you have already paired your Ecowitt device with the WS View Android/iOS app from\n  Ecowitt.\n\nFirst, install `ecowitt2mqtt` via `pip`:\n\n```bash\n$ pip install ecowitt2mqtt\n```\n\nThen, shift over to the WS View app on your Android/iOS device. While viewing your\ndevice in the app, select `Weather Services`:\n\n![Select Weather Services](https://raw.githubusercontent.com/bachya/ecowitt2mqtt/dev/assets/1-weather-services.jpeg?raw=true)\n\nPress `Next` until you reach the `Customized` screen:\n\n![The Customized screen in the WS View app](https://raw.githubusercontent.com/bachya/ecowitt2mqtt/dev/assets/2-customized.jpeg?raw=true)\n\nFill out the form with these values and tap `Save`:\n\n* `Protocol Type Same As`: `Ecowitt`\n* `Server IP / Hostname`: the IP address/hostname of the device running `ecowitt2mqtt`\n* `Path`: `/data/report` (note that unlike the default in the WS View App, there shouldn\'t\n  be a trailing slash)\n* `Port`: `8080` (the default port on which `ecowitt2mqtt` is served)\n* `Upload Interval`: `60` (change this to alter the frequency with which data is published)\n\nThen, on the machine where you installed `ecowitt2mqtt`, run it:\n\n```bash\n$ ecowitt2mqtt \\\n    --mqtt-broker=192.168.1.101 \\\n    --mqtt-username=user \\\n    --mqtt-password=password \\\n    --mqtt-topic=ecowitt2mqtt/device_1\n```\n\nWithin the `Upload Interval`, data should begin to appear in the MQTT broker.\n\n# Advanced Usage\n\n## Command Line Interface\n\nThe `ecowitt2mqtt` executable contains several configurable parameters:\n```\nusage: ecowitt2mqtt [-h] --mqtt-broker MQTT_BROKER --mqtt-topic MQTT_TOPIC [--mqtt-port MQTT_PORT]\n                    [--mqtt-username MQTT_USERNAME] [--mqtt-password MQTT_PASSWORD]\n                    [--endpoint ENDPOINT] [--port PORT] [-l LOG_LEVEL]\n\nSend data from Ecowitt devices to an MQTT broker\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -l LOG_LEVEL, --log-level LOG_LEVEL\n                        The logging level (default: INFO)\n  --mqtt-broker MQTT_BROKER\n                        The hostname or IP address of the MQTT broker\n  --mqtt-port MQTT_PORT\n                        The port of the MQTT broker (default: 1883)\n  --mqtt-username MQTT_USERNAME\n                        The username to use with the MQTT broker (default: None)\n  --mqtt-password MQTT_PASSWORD\n                        The password to use with the MQTT broker (default: None)\n  --mqtt-topic MQTT_TOPIC\n                        The MQTT topic to publish the device\'s data to (default: ecowitt2mqtt/<ID>)\n  --hass-discovery      Publish data in the Home Assistant MQTT Discovery format\n  --hass-discovery-prefix HASS_DISCOVERY_PREFIX\n                        The Home Assistant discovery prefix to use (default: homeassistant)\n  --endpoint ENDPOINT   The relative endpoint/path to serve the web app on (default: /data/report)\n  --port PORT           The port to serve the web app on (default: 8080)\n  --raw-data            Return raw data (don\'t attempt to translate any values)\n  --input-unit-system INPUT_UNIT_SYSTEM\n                        The input unit system used by the device (default: imperial)\n  --output-unit-system OUTPUT_UNIT_SYSTEM\n                        The unit system to use in output (default: imperial)\n```\n## Unit Systems\n\n`ecowitt2mqtt` allows you to specify both the input and output unit systems for a device.\nThis is fairly self-explanatory, but take care to use an `--input-unit-system` that is\nconsistent with what your device provides (otherwise, your data will be way off).\n\n## Raw Data\n\nIn some cases, it may be preferable to prevent `ecowitt2mqtt` from doing any data\ntranslation (converting values to a new unit system, changing binary values – such as\nmight be used by a battery – into "friendly" values, etc.). Passing the `--raw-data` flag\nwill accomplish this: data will flow directly from the Ecowitt device to the MQTT broker\nas-is.\n\nNote that the `--raw-data` flag supersedes any that might cause data translation (such as\n`--input-unit-system` or `--output-unit-system`).\n\n## Home Assistant MQTT Discovery\n\n[Home Assistant](https://home-assistant.io) users can quickly add entities from an\nEcowitt device by using\n[MQTT Discovery](https://www.home-assistant.io/docs/mqtt/discovery/).\n\nOnce Home Assistant is configured to accept MQTT Discovery, `ecowitt2mqtt` simply needs\nthe `--hass-discovery` flag:\n\n```bash\n$ ecowitt2mqtt \\\n    --mqtt-broker=192.168.1.101 \\\n    --mqtt-username=user \\\n    --mqtt-password=password \\\n    --hass-discovery\n```\n\nNote that if both `--hass-discovery` and `--mqtt-topic` are provided, `--hass-discovery` will\nwin out.\n\n## Running in the Background\n\n`ecowitt2mqtt` doesn\'t, itself, provide any sort of daemonization mechanism. The suggested\nroute is to use a different application.\n\n### `supervisord`\n\nAn example `supervisord` configuration file might look like this:\n\n```\n[supervisord]\nnodaemon=true\nloglevel=info\nuser=root\n\n[program:ecowitt2mqtt]\ncommand=ecowitt2mqtt --mqtt-broker=192.168.1.101 --mqtt-username=user --mqtt-password=password\nstdout_logfile=/dev/stdout\nstdout_logfile_maxbytes=0\nredirect_stderr=true\n```\n\n### `systemd`\n\nAn example `systemd` service file in `/etc/systemd/system` might look like this:\n\n```\n[Unit]\nDescription=ECOWITT2MQTT daemon\nAfter=network.target\n \n[Service]\nType=notify\nExecStart=ecowitt2mqtt --mqtt-broker=192.168.1.101 --mqtt-username=user --mqtt-password=password\nExecReload=kill -HUP $MAINPID\nKillMode=process\nRestart=on-failure\nRestartSec=5s\n \n[Install]\nWantedBy=multi-user.target\n```\n\nTo enable the service:\n\n```bash\n$ systemctl enable ecowitt2mqtt\n```\n\n## Docker\n\nThe library is available via a Docker image\n([`bachya/ecowitt2mqtt`](https://hub.docker.com/r/bachya/ecowitt2mqtt)). It is configured\nby a handful of environment variables that correspond to the command line parameters\nlisted above:\n\n* `LOG_LEVEL:` the log level to use (default: `INFO`)\n* `MQTT_BROKER:` the hostname or IP address of the MQTT broker\n* `MQTT_PORT:` the port of the MQTT broker (default: `1883`)\n* `MQTT_PASSWORD:` the password to use with the MQTT broker (default: `None`)\n* `MQTT_USERNAME:` the password to use with the MQTT broker (default: `None`)\n* `MQTT_TOPIC:` the MQTT topic to publish the device\'s data to\n* `HASS_DISCOVERY`: whether to use Home Assistant MQTT Discovery (default: `false`)\n* `HASS_DISCOVERY_PREFIX`: the topic prefix to use for Home Assistant MQTT Discovery\n  (default: `homeassistant`)\n* `ENDPOINT:` the relative endpoint/path to serve the web app on (default: `/data/report`)\n* `PORT:` the port to serve the web app on (default: `8080`)\n* `INPUT_UNIT_SYSTEM`: the input unit system to use (`imperial` or `metric`) (default: `imperial`)\n* `OUTPUT_UNIT_SYSTEM`: the output unit system to use (`imperial` or `metric`) (default: `imperial`)\n\nRunning the image is straightforward:\n\n```\ndocker run -it \\\n    -e MQTT_BROKER=192.168.1.101 \\\n    -e MQTT_USERNAME=user \\\n    -e MQTT_PASSWORD=password \\\n    -p 8080:8080 \\\n    bachya/ecowitt2mqtt:latest\n```\n\nNote the value of the `-p` flag: you must expose the port defined by the `PORT`\nenvironment variable. In the example above, the default port (`8080`) is used and is\nexposed via the same port on the host.\n\n[`docker-compose`](https://docs.docker.com/compose/) users can find an example\nconfiguration file at\n[`docker-compose.dev.yml`](https://github.com/bachya/ecowitt2mqtt/blob/dev/docker-compose.dev.yml).\nNote that this is intended to be a dev environment for quickly testing the repo itself;\nin production, you should refer to one of the\n[Docker Hub](https://hub.docker.com/r/bachya/ecowitt2mqtt) images.\n\n# Contributing\n\n1. [Check for open features/bugs](https://github.com/bachya/ecowitt2mqtt/issues)\nor [initiate a discussion on one](https://github.com/bachya/ecowitt2mqtt/issues/new).\n2. [Fork the repository](https://github.com/bachya/ecowitt2mqtt/fork).\n3. (_optional, but highly recommended_) Create a virtual environment: `python3 -m venv .venv`\n4. (_optional, but highly recommended_) Enter the virtual environment: `source ./.venv/bin/activate`\n5. Install the dev environment: `script/setup`\n6. Code your new feature or bug fix.\n7. Update `README.md` with any new documentation.\n8. Add yourself to `AUTHORS.md`.\n9. Submit a pull request!\n',
    'author': 'Aaron Bach',
    'author_email': 'bachya1208@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bachya/ecowitt2mqtt',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""Define constants."""
import logging

LOGGER = logging.getLogger(__package__)

# Glob data points:
DATA_POINT_GLOB_BAROM = "barom"
DATA_POINT_GLOB_BATT = "batt"
DATA_POINT_GLOB_BATT_BINARY = "batt_binary"
DATA_POINT_GLOB_GUST = "gust"
DATA_POINT_GLOB_HUMIDITY = "humidity"
DATA_POINT_GLOB_MOISTURE = "moisture"
DATA_POINT_GLOB_RAIN = "rain"
DATA_POINT_GLOB_TEMP = "temp"
DATA_POINT_GLOB_WIND = "wind"

# Specific data points:
DATA_POINT_CO2 = "co2"
DATA_POINT_DEWPOINT = "dewpoint"
DATA_POINT_FEELSLIKE = "feelslike"
DATA_POINT_HEATINDEX = "heatindex"
DATA_POINT_HUMIDITY = "humidity"
DATA_POINT_PM25 = "pm25"
DATA_POINT_PM25_24H = "pm25_24h"
DATA_POINT_SOLARRADIATION = "solarradiation"
DATA_POINT_SOLARRADIATION_LUX = "solarradiation_lux"
DATA_POINT_SOLARRADIATION_PERCEIVED = "solarradiation_perceived"
DATA_POINT_TEMPF = "tempf"
DATA_POINT_UV = "uv"
DATA_POINT_WINDCHILL = "windchill"
DATA_POINT_WINDDIR = "winddir"
DATA_POINT_WINDSPEEDMPH = "windspeedmph"

# Unit systems:
UNIT_SYSTEM_IMPERIAL = "imperial"
UNIT_SYSTEM_METRIC = "metric"
